import pytest


@pytest.fixture(scope="session")
def root_page(no):
    return no.pages.get("878d628488d94894ab14f9b872cd6870")


@pytest.fixture(scope="session")
def page_some_texts(no):
    return no.pages.get("82ee5677402f44819a5da3302273400a")


@pytest.fixture(scope="session")
def little_database(no):
    return no.databases.get("0e9539099cff456d89e44684d6b6c701")


@pytest.fixture(scope="session")
def database_for_updates(no):
    return no.databases.get("bf6ee5f75f99433a9d65132c05b42958")


@pytest.fixture(scope="session")
def page_for_pages(no):
    return no.pages.get("1bc86cc1d6f24362a6c40c2c89b423cc")


@pytest.fixture(scope="session")
def page_for_updates(no):
    return no.pages.get("36223246a20e42df8f9b354ed1f11d75")



import pytest

from pytion import Notion, setup_logging
from tests.fixtures import *


token = "secret_b3p8YOa25T1h9AscYVox8c5XO7yH8gh0tn7q3cfnnd0"


@pytest.fixture(scope="session")
def no():
    setup_logging("warning")
    return Notion(token)



import logging
from typing import Optional, Union

import pytion.envs as envs
from pytion.api import Notion
from pytion.exceptions import *


def setup_logging(
        level: Union[int, str] = logging.INFO, to_console: bool = True, filename: Optional[str] = None
) -> None:
    """

    :param level:       "debug", "info", "warning", "error", "critical" or `logging.INFO` etc.
    :param to_console:  True/False to output to stdout
    :param filename:    filename to put logs into file. or None
    :return:
    """
    log_levels = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }
    if isinstance(level, str):
        if level not in log_levels:
            raise ValueError("Invalid log level")
        level = log_levels[level]
    logger = logging.getLogger(__name__)
    logger.handlers.clear()
    logger.addHandler(logging.NullHandler())
    logger.setLevel(level)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    if to_console:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    if filename:
        file_handler = logging.FileHandler(filename)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)


setup_logging(level=envs.LOGGING_BASE_LEVEL, to_console=envs.LOGGING_TO_CONSOLE, filename=envs.LOGGING_FILE)



from datetime import datetime

import pytest

from pytion.models import Page, Block, Database, User, RichTextArray
from pytion.models import BlockArray, PropertyValue, PageArray, LinkTo, Property
from pytion import InvalidRequestURL, ObjectNotFound, ValidationError


def test_notion(no):
    assert no.version == "2022-02-22"
    assert no.session.base == "https://api.notion.com/v1/"


class TestElement:
    def test_get__page(self, root_page):
        page = root_page
        assert isinstance(page.obj, Page), "get of .pages. must return Page object"
        assert page.obj.id == "878d628488d94894ab14f9b872cd6870"
        assert str(page.obj.title) == "Pytion Tests"

    def test_get__block(self, no):
        block = no.blocks.get("878d628488d94894ab14f9b872cd6870")  # root page
        assert isinstance(block.obj, Block)
        assert block.obj.id == "878d628488d94894ab14f9b872cd6870"
        assert block.obj.text == "Pytion Tests"

    def test_get__database(self, no):
        database = no.databases.get("0e9539099cff456d89e44684d6b6c701")  # Little Database
        assert isinstance(database.obj, Database)
        assert database.obj.id == "0e9539099cff456d89e44684d6b6c701"
        assert str(database.obj.title) == "Little Database"

    def test_get__user(self, no):
        user = no.users.get("01c67faf3aba45ffaa022407f87c86a5")
        assert isinstance(user.obj, User)
        assert user.obj.id == "01c67faf3aba45ffaa022407f87c86a5"
        assert user.obj.name == "Yegor Gomzin"

    def test_get__bad_url(self, no):
        with pytest.raises(InvalidRequestURL):
            no.page.get("878d628488d94894ab14f9b872cd6870")  # root page

    def test_get__bad_id(self, no):
        with pytest.raises(ObjectNotFound):
            no.pages.get("878d628488d94894ab14f9b872cd6872")  # random id

    def test_get__after_path_page(self, no):
        blocks = no.blocks.get("82ee5677402f44819a5da3302273400a", _after_path="children")  # Page with some texts
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 3
        assert isinstance(blocks.obj[0], Block)
        blocks = no.blocks.get("82ee5677402f44819a5da3302273400a", _after_path="children", limit=2)  # Page with some te
        assert len(blocks.obj) == 2

    def test_get__after_path_block(self, no):
        blocks = no.blocks.get("8a920ba7dc1d4961811e5c82b28028ed", _after_path="children")  # Hello! How are you?
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 3
        assert isinstance(blocks.obj[0], Block)

    def test_get_parent__block(self, no):
        something = no.blocks.get_parent("9b2026c3a0cb45fc8cee330142d60f3a")  # I'm fine!
        assert something is None, "Blocks have not any parent"

    def test_get_parent__page(self, no):
        parent_page_block = no.pages.get_parent("82ee5677402f44819a5da3302273400a")  # Page with some texts
        assert isinstance(parent_page_block.obj, Block)
        assert parent_page_block.obj.id == "878d628488d94894ab14f9b872cd6870"
        assert parent_page_block.obj.text == "Pytion Tests"

    def test_get_parent__database(self, no):
        parent_page_block = no.pages.get_parent("82ee5677402f44819a5da3302273400a")  # Little Database
        assert isinstance(parent_page_block.obj, Block)
        assert parent_page_block.obj.id == "878d628488d94894ab14f9b872cd6870"
        assert parent_page_block.obj.text == "Pytion Tests"

    def test_get_parent__user(self, no):
        something = no.users.get_parent("01c67faf3aba45ffaa022407f87c86a5")
        assert something is None, "User object has not any parent"

    def test_get_parent__block_obj(self, no):
        block = no.blocks.get("8a920ba7dc1d4961811e5c82b28028ed")  # Hello! How are you?
        assert block.get_parent() is None, "Blocks have not any parent"

    def test_get_parent__page_obj(self, no):  # Database is the parent of this page
        page = no.pages.get("b85877eaf7bf4245a8c5218055eeb81f")  # Parent testing page
        database = page.get_parent()
        assert isinstance(database.obj, Database)
        assert database.obj.id == "0e9539099cff456d89e44684d6b6c701"
        assert str(database.obj.title) == "Little Database"

    def test_get_parent__database_obj(self, little_database):
        database = little_database  # Little Database
        parent_page_block = database.get_parent()
        assert isinstance(parent_page_block.obj, Block)
        assert parent_page_block.obj.id == "878d628488d94894ab14f9b872cd6870"
        assert parent_page_block.obj.text == "Pytion Tests"

    def test_get_parent__child_page(self, no):
        child_page = no.blocks.get("878d628488d94894ab14f9b872cd6870")  # root page
        page = child_page.get_parent()
        assert isinstance(page.obj, Page)
        assert page.obj.id == "878d628488d94894ab14f9b872cd6870"
        assert str(page.obj.title) == "Pytion Tests"

    def test_get_parent__child_database(self, no):
        child_database = no.blocks.get("0e9539099cff456d89e44684d6b6c701")  # Little Database
        database = child_database.get_parent()
        assert isinstance(database.obj, Database)
        assert database.obj.id == "0e9539099cff456d89e44684d6b6c701"
        assert str(database.obj.title) == "Little Database"

    def test_get_parent__workspace(self, root_page):
        workspace = root_page.get_parent()
        assert workspace is None

    def test_get_block_children__page_id(self, no):
        blocks = no.pages.get_block_children("82ee5677402f44819a5da3302273400a")  # Page with some texts
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 3
        assert isinstance(blocks.obj[0], Block)

    def test_get_block_children__page_obj(self, page_some_texts):
        page = page_some_texts  # Page with some texts
        blocks = page.get_block_children()
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 3
        assert isinstance(blocks.obj[0], Block)

    def test_get_block_children__block_id(self, no):
        blocks = no.blocks.get_block_children("8a920ba7dc1d4961811e5c82b28028ed")  # Hello! How are you?
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 3
        assert isinstance(blocks.obj[0], Block)

    def test_get_block_children__block_obj(self, no):
        block = no.blocks.get("8a920ba7dc1d4961811e5c82b28028ed")  # Hello! How are you?
        blocks = block.get_block_children()
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 3
        assert isinstance(blocks.obj[0], Block)

    def test_get_block_children__database_id(self, no):
        something = no.databases.get_block_children("0e9539099cff456d89e44684d6b6c701")  # Little Database
        assert something is None, "Database has no children"

    def test_get_block_children__database_obj(self, little_database):
        database = little_database  # Little Database
        something = database.get_block_children()
        assert something is None, "Database has no children"

    def test_get_block_children__child_database(self, no):
        database_block = no.blocks.get("0e9539099cff456d89e44684d6b6c701")  # Little Database
        database = database_block.get_block_children()
        assert isinstance(database.obj, Database)
        assert database.obj.id == "0e9539099cff456d89e44684d6b6c701"
        assert str(database.obj.title) == "Little Database"

    def test_get_block_children_recursive__page_id(self, no):
        blocks = no.pages.get_block_children_recursive("82ee5677402f44819a5da3302273400a")  # Page with some texts
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 6
        assert isinstance(blocks.obj[0], Block)

    def test_get_block_children_recursive__page_obj(self, page_some_texts):
        page = page_some_texts  # Page with some texts
        blocks = page.get_block_children_recursive()
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 6
        assert isinstance(blocks.obj[0], Block)

    def test_get_block_children_recursive__block_id(self, no):
        blocks = no.blocks.get_block_children_recursive("8a920ba7dc1d4961811e5c82b28028ed")  # Hello! How are you?
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 3
        assert isinstance(blocks.obj[0], Block)

    def test_get_block_children_recursive__block_obj(self, no):
        block = no.blocks.get("8a920ba7dc1d4961811e5c82b28028ed")  # Hello! How are you?
        blocks = block.get_block_children_recursive()
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 3
        assert isinstance(blocks.obj[0], Block)

    def test_get_block_children_recursive__database_id(self, no):
        something = no.databases.get_block_children_recursive("0e9539099cff456d89e44684d6b6c701")  # Little Database
        assert something is None, "Database has no children"

    def test_get_block_children_recursive__database_obj(self, little_database):
        database = little_database  # Little Database
        something = database.get_block_children_recursive()
        assert something is None, "Database has no children"

    def test_get_block_children_recursive__child_database(self, no):
        database_block = no.blocks.get("0e9539099cff456d89e44684d6b6c701")  # Little Database
        database = database_block.get_block_children_recursive()
        assert isinstance(database.obj, Database)
        assert database.obj.id == "0e9539099cff456d89e44684d6b6c701"
        assert str(database.obj.title) == "Little Database"

    def test_get_block_children_recursive__force(self, no):
        block = no.blocks.get("8a920ba7dc1d4961811e5c82b28028ed")  # Hello! How are you?
        blocks = block.get_block_children_recursive(force=True)
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 5
        assert isinstance(blocks.obj[0], Block)

    def test_get_block_children_recursive__max_depth(self, page_some_texts):
        page = page_some_texts  # Page with some texts
        blocks = page.get_block_children_recursive(force=True, max_depth=2)
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 7
        assert isinstance(blocks.obj[0], Block)

    def test_get_page_property__page_id(self, no):
        p = no.pages.get_page_property("%7Dma%3F", "b85877eaf7bf4245a8c5218055eeb81f")
        assert isinstance(p.obj, PropertyValue)
        assert len(p.obj.value) == 2
        assert p.obj.type == "multi_select"

    def test_get_page_property__page_obj(self, no):
        page = no.pages.get("b85877eaf7bf4245a8c5218055eeb81f")
        p = page.get_page_property("%7Dma%3F")
        assert isinstance(p.obj, PropertyValue)
        assert len(p.obj.value) == 2
        assert p.obj.type == "multi_select"

    def test_get_page_property__bad_id(self, no):
        with pytest.raises(ValidationError):
            no.pages.get_page_property("%7Dma%3A", "b85877eaf7bf4245a8c5218055eeb81f")

    def test_get_page_property__bad_page(self, no):
        with pytest.raises(ObjectNotFound):
            no.pages.get_page_property("%7Dma%3F", "b85877eaf7bf4245a8c5218055eeb81a")

    def test_db_query__id(self, no):
        pages = no.databases.db_query("0e9539099cff456d89e44684d6b6c701")  # Little Database
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 4
        assert str(pages.obj[0].title) == ""

    def test_db_query__obj(self, little_database):
        pages = little_database.db_query(limit=3)
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 3
        assert str(pages.obj[0].title) == ""

    def test_db_filter__title_ez(self, little_database):
        pages = little_database.db_filter("testing page")
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 1
        assert str(pages.obj[0].title) == "Parent testing page"

    def test_db_filter__not_contain(self, little_database):
        pages = little_database.db_filter("testing page", condition="does_not_contain")
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 3
        assert str(pages.obj[0].title) == ""

    def test_db_filter__ends_with(self, little_database):
        pages = little_database.db_filter(
            property_name="Name", property_type="title", value="what?", condition="ends_with"
        )
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 1
        assert str(pages.obj[0].title) == "wait, what?"

    def test_db_filter__is_empty(self, little_database):
        pages = little_database.db_filter(property_name="title", property_type="title", condition="is_empty")
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 1
        assert str(pages.obj[0].title) == ""

    def test_db_filter__greater_than(self, little_database):
        pages = little_database.db_filter(
            property_name="Digit", property_type="number", condition="greater_than", value="1"
        )
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 2
        assert str(pages.obj[0].title) == "We are best friends, body"

    def test_db_filter__checkbox(self, little_database):
        pages = little_database.db_filter(property_name="Done", property_type="checkbox")
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 1
        assert str(pages.obj[0].title) == "We are best friends, body"

    def test_db_filter__contains_tag(self, little_database):
        pages = little_database.db_filter(
            property_name="Tags", property_type="multi_select", value="tag1"
        )
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 2
        assert str(pages.obj[0].title) == ""

    def test_db_filter__notcontains_tag(self, little_database):
        pages = little_database.db_filter(
            property_name="Tags", property_type="multi_select", value="tag2", condition="does_not_contain"
        )
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 3
        assert str(pages.obj[0].title) == ""

    def test_db_filter__no_tags(self, little_database):
        pages = little_database.db_filter(
            property_name="Tags", property_type="multi_select", condition="is_empty"
        )
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 1
        assert str(pages.obj[0].title) == "wait, what?"

    def test_db_filter__tag_property_obj(self, no, little_database):
        page = no.pages.get("c2fc6b3dc3d244e9be2a3d28b26082bf")  # Untitled
        my_prop = page.obj.properties["Tags"]
        pages = little_database.db_filter(property_obj=my_prop)
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 2
        assert str(pages.obj[0].title) == ""

    def test_db_filter__without_filter(self, little_database):
        pages = little_database.db_filter("")
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 4
        assert str(pages.obj[0].title) == ""

    def test_db_filter__date_after(self, little_database):
        pages = little_database.db_filter(
            property_name="created", property_type="date", condition="on_or_after", value="2022-04-22"
        )
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 2
        assert str(pages.obj[0].title) == ""

    def test_db_filter__date_next_year(self, little_database):
        pages = little_database.db_filter(
            property_name="created", property_type="created_time", condition="next_year", value="2022-04-22"
        )
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 0

    def test_db_filter__sort_desc(self, little_database):
        pages = little_database.db_filter("", descending="created")
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 4
        assert str(pages.obj[0].title) == ""
        assert str(pages.obj[2].title) == "We are best friends, body"

    def test_db_filter__sort_asc(self, little_database):
        pages = little_database.db_filter("", ascending="Digit")
        assert isinstance(pages.obj, PageArray)
        assert len(pages.obj) == 4
        assert str(pages.obj[0].title) == "wait, what?"
        assert str(pages.obj[2].title) == "We are best friends, body"

    def test_db_create(self, no):
        parent = LinkTo.create(page_id="2dff77eb43d44ce097ffb421499f82aa")  # Page for creating databases
        properties = {
            "Name": Property.create("title"),
            "Digit": Property.create("number"),
            "Status": Property.create("select"),
        }
        title = "DB 1"
        database = no.databases.db_create(parent=parent, properties=properties, title=title)
        assert isinstance(database.obj, Database)
        assert str(database.obj.title) == title
        assert "Status" in database.obj.properties

        # Delete database manually. There is no way to delete a database by API

    def test_db_update__rename_title(self, database_for_updates):
        old_title = str(database_for_updates.obj.title)
        new_title = "1 BDU"
        database = database_for_updates.db_update(title=new_title)
        assert isinstance(database.obj, Database)
        assert str(database.obj.title) == new_title

        old_database = database_for_updates.db_update(title=old_title)
        assert str(old_database.obj.title) == old_title

    def test_db_update__rename_prop(self, database_for_updates):
        properties = {"Name": Property.create(type_="title", name="Subject")}
        database = database_for_updates.db_update(properties=properties)
        assert "Subject" in database.obj.properties

        title_property = database.obj.properties["Subject"]
        title_property.name = "Name"
        old_properties = {title_property.id: title_property}
        old_database = database.db_update(properties=old_properties)
        assert "Name" in old_database.obj.properties

    def test_db_update__retype_prop(self, database_for_updates):
        properties = {"Tags": Property.create("select")}
        database = database_for_updates.db_update(properties=properties)
        assert database.obj.properties["Tags"].type == "select"

        old_properties = {"Tags": Property.create("multi_select")}
        old_database = database_for_updates.db_update(properties=old_properties)
        assert old_database.obj.properties["Tags"].type == "multi_select"

    def test_db_update__create_delete_prop(self, database_for_updates):
        properties = {"New property": Property.create("checkbox")}
        database = database_for_updates.db_update(properties=properties)
        assert "New property" in database.obj.properties
        assert database.obj.properties["New property"].type == "checkbox"

        properties["New property"] = Property.create(None)
        database = database.db_update(properties=properties)
        assert "New property" not in database.obj.properties

    def test_page_create__into_page(self, no, page_for_pages):
        parent = LinkTo(from_object=page_for_pages.obj)
        page = no.pages.page_create(parent=parent, title="Page 1")
        assert isinstance(page.obj, Page)
        assert str(page.obj.title) == "Page 1"
        # delete section
        delete_page = page.page_update(archived=True)
        assert delete_page.obj.archived is True

    def test_page_create__into_database(self, no):
        parent = LinkTo.create(database_id="35f50aa293964b0d93e09338bc980e2e")  # Database for creating pages
        page = no.pages.page_create(parent=parent, title="Page 2")
        assert isinstance(page.obj, Page)
        assert str(page.obj.title) == "Page 2"
        # delete section
        delete_page = page.page_update(archived=True)
        assert delete_page.obj.archived is True

    def test_page_create__into_database_props(self, no):
        parent = LinkTo.create(database_id="35f50aa293964b0d93e09338bc980e2e")  # Database for creating pages
        props = {
            "Tags": PropertyValue.create(type_="multi_select", value=["tag1", "tag2"]),
            "done": PropertyValue.create("checkbox", True),
            "when": PropertyValue.create("date", datetime.now()),
        }
        page = no.pages.page_create(parent=parent, properties=props, title="Page 3")
        assert isinstance(page.obj, Page)
        assert str(page.obj.title) == "Page 3"
        # delete section
        delete_page = page.page_update(archived=True)
        assert delete_page.obj.archived is True

    def test_page_create__with_children(self, no, page_for_pages):
        parent = LinkTo(from_object=page_for_pages.obj)
        child = Block.create("Hello, World!")
        page = no.pages.page_create(parent=parent, title="Page 4", children=[child])
        assert isinstance(page.obj, Page)
        assert str(page.obj.title) == "Page 4"
        blocks = page.get_block_children()
        assert isinstance(blocks.obj, BlockArray)
        assert len(blocks.obj) == 1
        assert isinstance(blocks.obj[0], Block)
        assert str(blocks.obj[0].text) == "Hello, World!"
        # delete section
        delete_page = page.page_update(archived=True)
        assert delete_page.obj.archived is True

    def test_page_create__from_obj(self, no, page_for_pages):
        parent = LinkTo(from_object=page_for_pages.obj)
        page_obj = Page.create(parent=parent, title="Page 5")
        page = no.pages.page_create(page_obj=page_obj)
        assert isinstance(page.obj, Page)
        assert str(page.obj.title) == "Page 5"
        # delete section
        delete_page = page.page_update(archived=True)
        assert delete_page.obj.archived is True

    def test_page_update__rename(self, page_for_updates):
        old_name = str(page_for_updates.obj.title)
        new_name = "Updating for page"
        page = page_for_updates.page_update(title=new_name)
        assert isinstance(page.obj, Page)
        assert str(page.obj.title) == new_name

        old_page = page.page_update(title=RichTextArray.create(old_name))
        assert str(old_page.obj.title) == old_name

    def test_page_update__change_props(self, page_for_updates):
        old_props = page_for_updates.obj.properties
        new_props = {
            "Tags": PropertyValue.create("multi_select", ["tag2"]),
            "done": PropertyValue.create("checkbox", False),
            "when": PropertyValue.create("date", datetime.now()),
        }
        page = page_for_updates.page_update(properties=new_props)
        assert isinstance(page.obj, Page)
        assert "tag2" in page.obj.properties["Tags"].value
        assert page.obj.properties["done"].value is False

        old_page = page.page_update(properties=old_props)
        assert "tag1" in old_page.obj.properties["Tags"].value
        assert old_page.obj.properties["done"].value is True

    def test_page_update__delete(self, page_for_updates):
        page = page_for_updates.page_update(archived=True)
        assert isinstance(page.obj, Page)
        assert page.obj.archived is True

        old_page = page.page_update(archived=False)
        assert old_page.obj.archived is False

    def test_block_update__rename(self, no):
        block = no.blocks.get("08326405c2924ccc929bd78ceb70a2f3")  # Paragraph block to update.
        old_name = str(block.obj.text)
        new_name = old_name + old_name
        new_block = block.block_update(new_text=new_name)
        assert isinstance(new_block.obj, Block)
        assert str(new_block.obj.text) == new_name

        old_block = new_block.block_update(new_text=old_name)
        assert str(old_block.obj.text) == old_name

    def test_block_update__delete(self, no):
        block = no.blocks.get("08326405c2924ccc929bd78ceb70a2f3")  # Paragraph block to update.
        new_block = block.block_update(archived=True)
        assert isinstance(new_block.obj, Block)
        assert new_block.obj.archived is True

        old_block = new_block.block_update(archived=False)
        assert old_block.obj.archived is False

    def test_block_append__single(self, no):
        page = no.pages.get("365985ab349149d7826035fd46858b3f")  # Page for creating blocks
        my_block = Block.create("Such wow!")
        blocks = page.block_append(block=my_block)
        assert isinstance(blocks.obj, BlockArray)

        blocks = page.get_block_children()
        assert len(blocks.obj) == 1
        assert str(blocks.obj[0].text) == "Such wow!"
        remove_block = blocks.from_object(blocks.obj[0])
        removed_block = remove_block.block_update(archived=True)
        assert removed_block.obj.archived is True

    def test_block_append__many(self, no):
        my_block1 = Block.create("Do simple tests", type_="to_do", checked=True)
        my_block2 = Block.create(
            "multiline:\n  code: text", type_="code", caption="is it YAML?", language="yaml"
        )
        blocks = no.blocks.block_append("365985ab349149d7826035fd46858b3f", blocks=[my_block1, my_block2])
        assert isinstance(blocks.obj, BlockArray)

        blocks = no.pages.get_block_children("365985ab349149d7826035fd46858b3f")
        assert len(blocks.obj) == 2
        assert str(blocks.obj[0].text) == "[x] Do simple tests"
        assert str(blocks.obj[1].caption) == "is it YAML?"
        assert blocks.obj[1].language == "yaml"

        for block in blocks.obj:
            removed_block = no.blocks.block_update(block.id, archived=True)
            assert removed_block.obj.archived is True

    def test_from_linkto__base(self, no):
        link = LinkTo.create(page_id="878d628488d94894ab14f9b872cd6870")
        page = no.pages.from_linkto(link)
        assert isinstance(page.obj, Block)
        assert page.obj.id == "878d628488d94894ab14f9b872cd6870"
        assert str(page.obj.text) == "Pytion Tests"

    def test_from_linkto__child(self, page_some_texts):
        child = page_some_texts.from_linkto(page_some_texts.obj.children)
        assert isinstance(child.obj, BlockArray)
        assert len(child.obj) == 3
        assert isinstance(child.obj[0], Block)

    def test_from_linkto__parent(self, no, little_database):
        page = no.pages.get("b85877eaf7bf4245a8c5218055eeb81f")  # Parent testing page
        parent = page.from_linkto(page.obj.parent)
        assert isinstance(parent.obj, Database)
        assert parent.obj.id == little_database.obj.id
        assert str(parent.obj.title) == str(little_database.obj.title)



# -*- coding: utf-8 -*-
from __future__ import annotations
from datetime import datetime
from typing import Optional, Dict, Union, List, Any
from collections.abc import MutableSequence


# I wanna use pydantic, but API provide variable names of property


class RichText(object):
    def __init__(self, **kwargs) -> None:
        self.plain_text: str = kwargs.get("plain_text")
        self.href: Optional[str] = kwargs.get("href")
        self.annotations: Dict[str, Union[bool, str]] = kwargs.get("annotations")
        # if not self.annotations:
        #     self._create_default_annotations()
        self.type: str = kwargs.get("type")
        if self.type == "mention":
            subtype = kwargs[self.type].get("type")
            if subtype == "user":
                self.data = User(**kwargs[self.type].get(subtype))
                self.plain_text = str(self.data)
            elif subtype == "page":
                sub_id = kwargs[self.type][subtype].get("id") if kwargs[self.type].get(subtype) else ""
                self.data = LinkTo.create(page=sub_id)
                if self.plain_text == "Untitled":
                    self.plain_text = repr(self.data)
                else:
                    self.plain_text = "LinkTo(" + self.plain_text + ")"
            elif subtype == "database":
                sub_id = kwargs[self.type][subtype].get("id") if kwargs[self.type].get(subtype) else ""
                self.data = LinkTo.create(database_id=sub_id)
                if self.plain_text == "Untitled":
                    self.plain_text = repr(self.data)
                else:
                    self.plain_text = "LinkTo(" + self.plain_text + ")"
            elif subtype == "date":
                self.data = {
                    "start": Model.format_iso_time(kwargs[self.type][subtype].get("start")),
                    "end": Model.format_iso_time(kwargs[self.type][subtype].get("end"))
                }
            elif subtype == "link_preview":
                self.plain_text = f"<{self.plain_text}>"
                self.data: Dict = kwargs[self.type]
        else:
            self.data: Dict = kwargs[self.type]

    def __str__(self):
        return str(self.plain_text)

    def __repr__(self):
        return f"RichText({self.plain_text})"

    def __bool__(self):
        return bool(self.plain_text)

    def _create_default_annotations(self):
        self.annotations = {
            "bold": False, "italic": False, "strikethrough": False,
            "underline": False, "code": False, "color": "default"
        }

    # def __len__(self):
    #     return len(self.plain_text)

    def get(self) -> Dict[str, Any]:
        """
        Text type supported only
        """
        return {
            "type": "text",
            "text": {"content": self.plain_text, "link": None},
            # "annotations": self.annotations,
            # "plain_text": self.plain_text,
        }


class RichTextArray(MutableSequence):
    def __init__(self, array: List[Dict]) -> None:
        self.array = [RichText(**rt) for rt in array]

    def __getitem__(self, item):
        return self.array[item]

    def __setitem__(self, key, value):
        self.array[key] = value

    def __delitem__(self, key):
        del self.array[key]

    def __len__(self):
        return len(self.array)

    def insert(self, index: int, value) -> None:
        self.array.insert(index, value)

    def __str__(self):
        return " ".join(str(rt) for rt in self)

    def __repr__(self):
        return f"RichTextArray({str(self)})"

    def __bool__(self):
        return any(map(bool, self.array))

    def __add__(self, another: Union[RichTextArray, str]):
        if isinstance(another, str):
            another = RichTextArray.create(another)
        self.array.extend(another)
        return self

    def get(self) -> List[Dict[str, Any]]:
        return [item.get() for item in self]

    @classmethod
    def create(cls, text: str):
        return cls([{"type": "text", "plain_text": text, "text": {}}])


class User(object):
    """
    The User object represents a user in a Notion workspace.
    """
    path = "users"

    def __init__(self, **kwargs) -> None:
        """
        Create an User object by providing dict from API.

        API attrs (from API docs):
        Mandatory:
        :param id: str
        :param object: str

        Optional:
        :param type: str
        :param name: str
        :param avatar_url: str

        Also Local attrs:
        :param raw: dict from API
        :param email: str if user is person
        """
        self.id = kwargs.get("id", "").replace("-", "")
        self.object = kwargs.get("object")  # user
        self.type = kwargs.get("type")
        self.name = kwargs.get("name")
        self.avatar_url = kwargs.get("avatar_url")
        if self.type == "person" and kwargs.get(self.type):
            self.email = kwargs[self.type].get("email")
        else:
            self.email = None
        self.raw = kwargs

    def __str__(self):
        if self.name and self.email:
            name = f"{self.name}({self.email})"
        else:
            name = self.name
        return name if name else self.id

    def __repr__(self):
        return f"User({self})"

    def get(self) -> Dict[str, str]:
        return {
            "object": self.object,
            "id": self.id
        }

    @classmethod
    def create(cls, id: str):
        return cls(object="user", id=id)


class Model(object):
    """
    :param id:
    :param object:
    :param created_time:
    :param last_edited_time:
    :param created_by:
    :param last_edited_by:
    :param raw:
    """

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id", "").replace("-", "")
        self.object = kwargs.get("object")
        self.created_time = self.format_iso_time(kwargs.get("created_time"))
        self.last_edited_time = self.format_iso_time(kwargs.get("last_edited_time"))
        self.created_by = User(**kwargs["created_by"]) if kwargs.get("created_by") else None
        self.last_edited_by = User(**kwargs["last_edited_by"]) if kwargs.get("last_edited_by") else None
        self.raw = kwargs

    @classmethod
    def format_iso_time(cls, time: str) -> Optional[datetime]:
        if not time:
            return None
        return datetime.fromisoformat(time.replace("Z", "+00:00"))


class Property(object):
    def __init__(self, data: Dict[str, str]):
        self.to_delete = True if data.get("type") is None else False
        self.id: str = data.get("id")
        self.type: str = data.get("type", "")
        self.name: str = data.get("name")
        self.raw = data

    def __str__(self):
        return self.name if self.name else self.type

    def __repr__(self):
        return f"Property({self})"

    def get(self) -> Optional[Dict[str, Dict]]:
        # property removing while patch
        if self.to_delete:
            return None
        # property renaming while patch
        data = {}
        if self.name:
            data["name"] = self.name
        # property retyping while patch
        if self.type:
            data[self.type] = {}
        return data

    @classmethod
    def create(cls, type_: Optional[str] = "", **kwargs):
        """
        Property Schema Object (watch docs)

        + addons:
        set type = `None` to delete this Property
        set param `name` to rename this Property
        """
        return cls({"type": type_, **kwargs})


class PropertyValue(Property):
    def __init__(self, data: Dict, name: str, **kwargs):
        super().__init__(data)
        # getting Paginated Properties (for retrieving property item)
        # *Pagination
        if data.get("object") and data["object"] == "list":
            if data.get("results"):
                self.type = data["results"][0].get("type")
                data[self.type] = [sub_dict.get(sub_dict.get("type")) for sub_dict in data["results"]]

        self.name = name
        self.value = None

        if self.type in ["title", "rich_text"]:
            if isinstance(data[self.type], list):
                self.value = RichTextArray(data[self.type])
            elif isinstance(data[self.type], RichTextArray):
                self.value = data[self.type]
            else:
                self.value = RichTextArray.create(data[self.type])

        if self.type == "number":
            self.value: Optional[int, float] = data["number"]

        if self.type == "select":
            if data["select"] and isinstance(data["select"], dict):
                self.value: Optional[str] = data["select"].get("name")
            elif data["select"] and isinstance(data["select"], str):
                self.value = data["select"]
            else:
                self.value = None

        if self.type == "multi_select":
            self.value: List[str] = [(v.get("name") if isinstance(v, dict) else v) for v in data["multi_select"]]

        if self.type == "checkbox":
            self.value: bool = data["checkbox"]

        if self.type == "date":
            if isinstance(data["date"], datetime):
                self.value = data["date"].isoformat()
                self.start = data["date"]
                self.end = None
            elif data["date"]:
                self.value: Optional[str] = data["date"].get("start")
                self.start: Optional[datetime] = Model.format_iso_time(data["date"].get("start"))
                self.end: Optional[datetime] = Model.format_iso_time(data["date"].get("end"))
            else:
                self.value = None
                self.start = None
                self.end = None

        if "time" in self.type:
            self.value: Optional[datetime] = Model.format_iso_time(data.get(self.type))

        if self.type == "formula":
            formula_type = data["formula"]["type"]
            if formula_type == "date":
                if data["formula"]["date"]:
                    self.value: str = data["formula"]["date"].get("start")
                    self.start: Optional[datetime] = Model.format_iso_time(data["formula"]["date"].get("start"))
                    self.end: Optional[datetime] = Model.format_iso_time(data["formula"]["date"].get("end"))
                else:
                    self.value = None
                    self.start = None
                    self.end = None
            else:
                self.value: Union[str, int, float, bool] = data["formula"][formula_type]

        if self.type == "created_by":
            self.value = User(**data.get(self.type))

        if self.type == "last_edited_by":
            self.value = User(**data.get(self.type))

        if self.type == "people":
            self.value = [user if isinstance(user, User) else User(**user) for user in data[self.type]]

        if self.type == "relation":
            self.value: List[str] = [item.get("id") for item in data["relation"]]

        if self.type == "rollup":
            rollup_type = data["rollup"]["type"]
            if rollup_type == "array":
                if len(data["rollup"]["array"]) == 0:
                    self.value = None
                elif len(data["rollup"]["array"]) == 1:
                    self.value = PropertyValue(data["rollup"]["array"][0], rollup_type)
                else:
                    self.value = [PropertyValue(element, rollup_type) for element in data["rollup"]["array"]]

            if rollup_type == "number":
                self.value: Optional[int, float] = data["rollup"]["number"]

            if rollup_type == "date":
                if data["rollup"]["date"]:
                    self.value: Optional[str] = data["rollup"]["date"].get("start")
                    self.start: Optional[datetime] = Model.format_iso_time(data["rollup"]["date"].get("start"))
                    self.end: Optional[datetime] = Model.format_iso_time(data["rollup"]["date"].get("end"))
                else:
                    self.value = None
                    self.start = None
                    self.end = None
            else:
                self.value = "unsupported rollup type"

        if self.type == "files":
            self.value = "unsupported"

        if self.type == "url":
            self.value: Optional[str] = data.get("url")

        if self.type == "email":
            self.value: Optional[str] = data.get("email")

        if self.type == "phone_number":
            self.value: Optional[str] = data.get("phone_number")

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f"{self.name}({self})"

    def get(self):
        # checkbox can not be `None`
        if self.type in ["checkbox"]:
            return {self.type: self.value}

        # empty values
        if not self.value:
            if self.type in ["multi_select", "relation", "rich_text", "people", "files"]:
                return {self.type: []}
            return {self.type: None}

        # RichTextArray
        if self.type in ["title", "rich_text"] and hasattr(self.value, "get"):
            return {self.type: self.value.get()}

        # simple values
        if self.type in ["number", "url", "email", "phone_number"]:
            return {self.type: self.value}

        # select type
        if self.type == "select":
            return {self.type: {"name": self.value}}

        # multi-select type
        if self.type == "multi_select":
            return {self.type: [{"name": tag} for tag in self.value]}

        # date type
        if self.type == "date" and hasattr(self, "start") and hasattr(self, "end"):
            with_time = True if self.start.hour or self.start.minute else False
            if self.start:
                start = self.start.astimezone().isoformat() if with_time else str(self.start.date())
            else:
                start = None
            if self.end:
                end = self.end.astimezone().isoformat() if with_time else str(self.end.date())
            else:
                end = None
            return {self.type: {"start": start, "end": end}}

        # people type
        if self.type == "people":
            return {self.type: [user.get() for user in self.value]}

        # unsupported types:
        if self.type in ["files", "relation"]:
            return {self.type: []}
        if self.type in ["created_time", "last_edited_by", "last_edited_time", "created_by"]:
            return None
        if self.type in ["formula", "rollup"]:
            return {self.type: {}}
        return None

    @classmethod
    def create(cls, type_: str = "", value: Any = None, **kwargs):
        """
        Property Value Object (watch docs)

        + addons:
        set type = `None` to delete this Property
        set param `name` to rename this Property
        """
        return cls({"type": type_, type_: value, **kwargs}, name="")


class Database(Model):
    object = "database"
    path = "databases"

    def __init__(self, **kwargs) -> None:
        """
        params from Model +
        :param cover:
        :param icon:
        :param title:
        :param properties:
        :param parent:
        :param url:
        """
        super().__init__(**kwargs)
        self.cover: Optional[Dict] = kwargs.get("cover")
        self.icon: Optional[Dict] = kwargs.get("icon")
        self.title = (kwargs.get("title")
                      if isinstance(kwargs["title"], RichTextArray) or not kwargs.get("title")
                      else RichTextArray(kwargs["title"])
                      )
        self.properties = {
            name: (value if isinstance(value, Property) else Property(value))
            for name, value in kwargs["properties"].items()
        }
        self.parent = kwargs["parent"] if isinstance(kwargs.get("parent"), LinkTo) else LinkTo(**kwargs["parent"])
        self.url: str = kwargs.get("url")

    def __str__(self):
        return str(self.title)

    def __repr__(self):
        return f"Database({self.title})"

    def get(self) -> Dict[str, Dict]:
        new_dict = {
            "parent": self.parent.get(),
            "properties": {name: value.get() for name, value in self.properties.items()}
        }
        if isinstance(self.title, RichTextArray):
            new_dict["title"] = self.title.get()
        return new_dict

    @classmethod
    def create(cls, parent: LinkTo, properties: Dict[str, Property], title: Optional[RichTextArray] = None, **kwargs):
        return cls(parent=parent, properties=properties, title=title, **kwargs)


class Page(Model):
    object = "page"
    path = "pages"

    def __init__(self, **kwargs) -> None:
        """
        params from Model +
        :param cover:
        :param icon:
        :param parent:
        :param archived:
        :param properties:
        :param url:
        """
        super().__init__(**kwargs)
        self.cover: Optional[Dict] = kwargs.get("cover")
        self.icon: Optional[Dict] = kwargs.get("icon")
        self.parent = kwargs["parent"] if isinstance(kwargs.get("parent"), LinkTo) else LinkTo(**kwargs["parent"])
        self.archived: bool = kwargs.get("archived")
        self.url: str = kwargs.get("url")
        self.children = kwargs["children"] if "children" in kwargs else LinkTo(block=self)
        self.properties = {
            name: (PropertyValue(data, name) if not isinstance(data, PropertyValue) else data)
            for name, data in kwargs["properties"].items()
        }
        for p in self.properties.values():
            if "title" in p.type:
                self.title = p.value
                break
        else:
            self.title = None

    def __str__(self):
        return str(self.title)

    def __repr__(self):
        return f"Page({self.title})"

    def get(self):
        new_dict = {
            "parent": self.parent.get(without_type=True),
            "icon": self.icon,  # optional
            "cover": self.cover,  # optional
            "properties": {name: p.get() for name, p in self.properties.items()},
        }
        if getattr(self, "children", None):
            new_dict["children"] = self.children.get()  # can not be None
        return new_dict

    @classmethod
    def create(
            cls, parent: LinkTo, properties: Optional[Dict[str, PropertyValue]] = None,
            title: Optional[RichTextArray, str] = None, children: Optional[BlockArray] = None, **kwargs
    ):
        if not properties:
            properties = {}
        if title:
            properties["title"] = PropertyValue.create("title", title)
        return cls(parent=parent, properties=properties, children=children, **kwargs)


class Block(Model):
    object = "block"
    path = "blocks"

    def __init__(self, **kwargs):
        """
        params from Model +
        :param has_children:
        :param type:
        :param archived:
        """
        super().__init__(**kwargs)
        self.type: str = kwargs.get("type")
        self.has_children: bool = kwargs.get("has_children")
        self.archived: bool = kwargs.get("archived")
        self.children = LinkTo(block=self)
        self._level = kwargs["level"] if kwargs.get("level") else 0
        self.create_mode: bool = kwargs["create_mode"] if "create_mode" in kwargs else False
        self.parent = None

        if self.create_mode:
            self.text = kwargs[self.type]
            if "checked" in kwargs:
                self.checked = kwargs["checked"]
            if "language" in kwargs:
                self.language = kwargs["language"]
            if "caption" in kwargs:
                self.caption = kwargs["caption"]
                if isinstance(self.caption, str):
                    self.caption = RichTextArray.create(self.caption)
            return

        if self.type == "paragraph":
            self.text = RichTextArray(kwargs[self.type].get("rich_text"))
            # Paragraph Block does not contain `children` attr (watch Docs)

        elif "heading" in self.type:
            indent = self.type.split("_")[-1]
            indent_num = int(indent) if indent.isdigit() else 0
            prefix = "#" * indent_num
            self.text = RichTextArray.create(prefix) + RichTextArray(kwargs[self.type].get("rich_text"))

        elif self.type == "callout":
            self.text = RichTextArray(kwargs[self.type].get("rich_text"))
            self.icon: Dict = kwargs[self.type].get("icon")
            # Callout Block does not contain `children` attr (watch Docs)

        elif self.type == "quote":
            self.text = RichTextArray.create("|") + RichTextArray(kwargs[self.type].get("rich_text"))
            # Quote Block does not contain `children` attr (watch Docs)

        elif "list_item" in self.type:
            self.text = RichTextArray.create("-") + RichTextArray(kwargs[self.type].get("rich_text"))
            # Block does not contain `children` attr (watch Docs)
            # Numbers does not support cause of lack of relativity

        elif self.type == "to_do":
            self.checked: bool = kwargs[self.type].get("checked")
            prefix = "[x]" if self.checked else "[ ]"
            self.text = RichTextArray.create(prefix) + RichTextArray(kwargs[self.type].get("rich_text"))
            # To-do Block does not contain `children` attr (watch Docs)

        elif self.type == "toggle":
            self.text = RichTextArray.create(">") + RichTextArray(kwargs[self.type].get("rich_text"))
            # Toggle Block does not contain `children` attr (watch Docs)

        elif self.type == "code":
            self.text = RichTextArray.create("```\n") + RichTextArray(kwargs[self.type].get("rich_text")) + "\n```"
            self.language: str = kwargs[self.type].get("language")
            self.caption = RichTextArray(kwargs[self.type].get("caption"))

        # when the block is child_page, parent will be the page object
        # when the block is child_database, parent AND children will be the database object
        elif "child" in self.type:
            self.text: str = kwargs[self.type].get("title")
            if self.type == "child_page":
                # self.children is already set
                self.parent = LinkTo(type="page", page=self.id)
            elif self.type == "child_database":
                # well yes. parent and children are the same. parent of this database will be the page of this block
                # and the database is children of this block
                self.parent = LinkTo.create(database_id=self.id)
                self.children = LinkTo.create(database_id=self.id)
                if not self.text:
                    self.text = repr(self.children)
            # page self.has_children is correct. checked.
            # database self.has_children is false.
            # database with custom source had no title!

        # hello, markdown
        elif self.type == "embed":
            self.caption = RichTextArray(kwargs[self.type].get("caption"))
            if self.caption:
                self.text = f'[{self.caption}]({kwargs[self.type].get("url")})'
            else:
                self.text = f'<{kwargs[self.type]["url"]}>' if kwargs[self.type].get("url") else "*Empty embed*"

        elif self.type == "image":
            self.caption = RichTextArray(kwargs[self.type].get("caption"))
            subtype = kwargs[self.type].get("type")
            if subtype == "file":
                # The file S3 URL will be valid for 1 hour
                self.expiry_time = Model.format_iso_time(kwargs[self.type][subtype].get("expiry_time"))
            else:
                self.expiry_time = None
            if subtype in ("file", "external"):
                if self.caption:
                    self.text = f'[{self.caption}]({kwargs[self.type][subtype].get("url")})'
                else:
                    self.text = f'<{kwargs[self.type][subtype].get("url")}>'
            else:
                self.text = "*Unknown image type*"

        elif self.type == "video":
            self.caption = RichTextArray(kwargs[self.type].get("caption"))
            subtype = kwargs[self.type].get("type")
            if subtype == "file":
                self.expiry_time = Model.format_iso_time(kwargs[self.type][subtype].get("expiry_time"))
            else:
                self.expiry_time = None
            if subtype in ("file", "external"):
                if self.caption:
                    self.text = f'[{self.caption}]({kwargs[self.type][subtype].get("url")})'
                else:
                    self.text = f'<{kwargs[self.type][subtype].get("url")}>'
            else:
                self.text = "*Unknown video type*"

        elif self.type == "file":
            self.caption = RichTextArray(kwargs[self.type].get("caption"))
            subtype = kwargs[self.type].get("type")
            if subtype == "file":
                self.expiry_time = Model.format_iso_time(kwargs[self.type][subtype].get("expiry_time"))
            else:
                self.expiry_time = None
            if subtype in ("file", "external"):
                if self.caption:
                    self.text = f'[{self.caption}]({kwargs[self.type][subtype].get("url")})'
                else:
                    self.text = f'<{kwargs[self.type][subtype].get("url")}>'
            else:
                self.text = "*Unknown file type*"

        elif self.type == "pdf":
            self.caption = RichTextArray(kwargs[self.type].get("caption"))
            subtype = kwargs[self.type].get("type")
            if subtype == "file":
                self.expiry_time = Model.format_iso_time(kwargs[self.type][subtype].get("expiry_time"))
            else:
                self.expiry_time = None
            if subtype in ("file", "external"):
                if self.caption:
                    self.text = f'[{self.caption}]({kwargs[self.type][subtype].get("url")})'
                else:
                    self.text = f'<{kwargs[self.type][subtype].get("url")}>'
            else:
                self.text = "*Unknown pdf type*"

        elif self.type == "breadcrumb":
            self.text = "*breadcrumb block*"

        elif self.type == "bookmark":
            self.caption = RichTextArray(kwargs[self.type].get("caption"))
            if self.caption:
                self.text = f'[{self.caption}]({kwargs[self.type].get("url")})'
            else:
                self.text = f'<{kwargs[self.type]["url"]}>' if kwargs[self.type].get("url") else "*Empty bookmark*"

        elif self.type == "link_preview":
            self.text = f'<{kwargs[self.type].get("url")}>'

        elif self.type == "link_to_page":
            self.link = LinkTo(**kwargs[self.type])
            self.text = repr(self.link)

        elif self.type == "equation":
            self.text: str = kwargs[self.type].get("expression")

        elif self.type == "divider":
            self.text = "---"

        elif self.type == "table_of_contents":
            self.text = "*Table of contents*"

        elif self.type == "template":
            self.text = RichTextArray.create("Template:") + RichTextArray(kwargs[self.type].get("rich_text"))

        elif self.type == "synced_block":
            synced_from = kwargs[self.type].get("synced_from")
            self.text = "*SYNCED BLOCK:*"
            self.synced_from = LinkTo(**synced_from) if synced_from else None

        elif self.type == "table":
            self.table_width: int = kwargs[self.type].get("table_width")
            self.text = f"*Table {self.table_width}xN:*"

        elif self.type == "table_row":
            cells = kwargs[self.type].get("cells")
            self.text = RichTextArray.create("|")
            for cell in cells:
                self.text += RichTextArray(cell) + "|"

        elif self.type == "unsupported":
            self.text = "*****"

        else:
            self.text = "*UNKNOWN_BLOCK_TYPE*"

    def __str__(self):
        return str(self.text)

    def __repr__(self):
        return f"Block({str(self.text)[:30]})"

    def get(self, with_object_type: bool = False):
        if self.type in [
            "paragraph", "quote", "heading_1", "heading_2", "heading_3", "to_do",
            "bulleted_list_item", "numbered_list_item", "toggle", "callout", "code", "child_database"
        ]:

            text = RichTextArray.create(self.text) if isinstance(self.text, str) else self.text
            new_dict = {self.type: {"rich_text": text.get()}}
            if self.type == "to_do" and hasattr(self, "checked"):
                new_dict[self.type]["checked"] = self.checked
            if self.type == "code":
                new_dict[self.type]["language"] = getattr(self, "language", "plain text")
                if hasattr(self, "caption"):
                    new_dict[self.type]["caption"] = self.caption.get()
            if self.type == "child_database":
                new_dict = {self.type: {"title": str(text)}}
            if with_object_type:
                new_dict["object"] = "block"
                new_dict["type"] = self.type
            return new_dict
        return None

    @classmethod
    def create(cls, text: str, type_: str = "paragraph", **kwargs):
        """
        :param text:   Block content
        :param type_:  Block types (API)
        :param kwargs:
            :kwargs param checked:  bool for to_do
            :kwargs param language: str for code
            :kwargs param caption:  str or RichTextArray for code
        :return:
        """
        new_dict = {
            "type": type_,
            type_: text,
        }
        return cls(**new_dict, create_mode=True, **kwargs)


class ElementArray(MutableSequence):
    class_map = {"page": Page, "database": Database, "block": Block}

    def __init__(self, array, create: bool = False):
        if create:
            self.array = array
            return

        self.array = []
        for ele in array:
            if ele.get("object") and ele["object"] in self.class_map:
                self.array.append(self.class_map[ele["object"]](**ele))

    def __getitem__(self, item):
        return self.array[item]

    def __setitem__(self, key, value):
        self.array[key] = value

    def __delitem__(self, key):
        del self.array[key]

    def __len__(self):
        return len(self.array)

    def insert(self, index: int, value) -> None:
        self.array.insert(index, value)

    def __str__(self):
        return "\n".join(str(b) for b in self)

    def __repr__(self):
        r = str(self)[:30].replace("\n", " ")
        return f"ElementArray({r})"


class BlockArray(ElementArray):
    def __str__(self):
        return "\n".join(b._level * "\t" + str(b) for b in self)

    def __repr__(self):
        r = str(self)[:30].replace("\n", " ")
        return f"BlockArray({r})"

    def get(self):
        return [b.get() for b in self]


class PageArray(ElementArray):
    def __repr__(self):
        r = str(self)[:30].replace("\n", " ")
        return f"PageArray({r})"


class LinkTo(object):
    """
    schema
    .type = `element_type`
    .id = `elementID`

    .get() - return API like style
    .create() - create in format `(page_id="123412341234")` or (database_id="13412341234")`
    """

    def __init__(
            self, block: Optional[Model] = None, from_object: Optional[Block, Page, Database] = None, **kwargs
    ):
        """
        Creates LinkTo object from API dict

        :param block: Block object can be provided instead other attrs. Internal usage.
        :param from_object: Any model object can be provided to create LinkTo to it.
        :param kwargs: API attrs. Internal usage.
        """
        if block:
            self.type = block.object
            self.id = block.id
            self.after_path = "children"
            self.uri = "blocks"
        # You can provide the object to create the LinkTo to it
        elif from_object:
            self.uri = from_object.path
            self.id = from_object.id
            if isinstance(from_object, Page):
                self.type = "page_id"
            elif isinstance(from_object, Database):
                self.type = "database_id"
            elif isinstance(from_object, Block):
                self.type = "block_id"
            elif isinstance(from_object, User):
                self.type = "user_id"
        else:
            self.type: str = kwargs.get("type")
            self.id: str = kwargs.get(self.type) if kwargs.get(self.type) else kwargs.get("id")
            self.after_path = ""
            if self.type == "page_id":
                self.uri = "blocks"
            elif self.type == "database_id":
                self.uri = "databases"
            # when type is set manually
            elif self.type == "page":
                self.uri = "pages"
            elif self.type == "block_id":
                self.uri = "blocks"
            elif self.type == "user_id":
                self.uri = "users"
            else:
                self.uri = None

        if isinstance(self.id, str):
            self.id = self.id.replace("-", "")

    def __str__(self):
        prefix = self.uri if self.uri else self.type
        if getattr(self, "after_path", ""):
            return f"{prefix}/{self.id}/{self.after_path}"
        return f"{prefix}/{self.id}"

    def __repr__(self):
        return f"LinkTo({self})"

    def get(self, without_type: bool = False):
        if without_type:
            return {self.type: self.id}
        return {"type": self.type, self.type: self.id}

    @classmethod
    def create(cls, **kwargs):
        """
        `.create(page_id="123412341234")`
        `.create(database_id="13412341234")`
        """
        for key, value in kwargs.items():
            return cls(type=key, id=value)



# -*- coding: utf-8 -*-

import logging


# Base URL (mandatory)
NOTION_URL = "https://api.notion.com/v1/"

# Access token (optional)
try:
    with open("token") as f:
        NOTION_SECRET = f.read()
except FileNotFoundError:
    NOTION_SECRET = None

# Current API Version (mandatory)
NOTION_VERSION = "2022-02-22"

# Logging settings (mandatory)
LOGGING_BASE_LEVEL = logging.WARNING
LOGGING_TO_CONSOLE = False
# set `None` to do not logging into file
LOGGING_FILE = None

# every resource has `object` property (type declaration)
# every resource has `id` property (UUIDv4)
# every property is in snake_case only
# temporal values - ISO 8601
# 2020-08-12T02:12:33.231Z
# 2020-08-12T02:12:33.231+00:00
# 2020-08-12

# empty strings DOES NOT supported. use `None` (python) or `null` (JSON)



# -*- coding: utf-8 -*-

import logging
from urllib.parse import urlencode
from typing import Dict, Optional, Any, Union
from datetime import datetime

import requests

import pytion.envs as envs
from pytion.models import Property, PropertyValue, User
from pytion.exceptions import find_response_error


logger = logging.getLogger(__name__)


class Filter(object):
    _filter_condition_types = [
        "rich_text", "number", "checkbox", "select", "multi_select", "date", "phone_number", "people", "title",
        "created_time", "last_edited_time", "phone_number"
    ]

    def __init__(
            self,
            property_name: Optional[str] = None,
            value: Optional[Any] = None,
            property_type: Optional[str] = None,
            condition: Optional[str] = None,
            raw: Optional[Dict] = None,
            property_obj: Optional[Union[Property, PropertyValue]] = None,
            **kwargs,
    ):
        if raw:
            self.filter = raw
            return
        if property_obj:
            if property_obj.id:
                self.property_name = property_obj.id
            else:
                self.property_name = property_obj.name
            if property_obj.type in ["title", "rich_text", "url", "email"]:
                self.property_type = "rich_text"
            elif "time" in property_obj.type:
                self.property_type = "date"
            else:
                self.property_type = property_obj.type
        else:
            self.property_type = property_type
            self.property_name = property_name

        if self.property_type not in self._filter_condition_types:
            raise ValueError(f"Allowed types {self.allowed_condition_types} ({property_type} is provided)")

        if self.property_type == "rich_text":
            self.condition = "contains" if not condition else condition
            self.value = str(value)
        elif self.property_type == "number":
            self.condition = "equals" if not condition else condition
            if "." in value:
                self.value = float(value)
            else:
                self.value = int(value)
        elif self.property_type == "checkbox":
            self.condition = "equals" if not condition else condition
            self.value = bool(value) if value else True
        elif self.property_type == "select":
            self.condition = "equals" if not condition else condition
            self.value = str(value)
        elif self.property_type == "multi_select":
            self.condition = "contains" if not condition else condition
            self.value = value[0] if isinstance(value, list) else str(value)
        elif self.property_type == "phone_number":
            self.condition = "contains" if not condition else condition
            self.value = str(value)
        elif self.property_type == "people":
            self.condition = "contains" if not condition else condition
            if isinstance(value, User):
                self.value = value.id
            else:
                self.value = str(value)
        elif self.property_type == "title":
            self.condition = "contains" if not condition else condition
            self.value = str(value)
        elif self.property_type == "date" or "_time" in self.property_type:
            self.condition = "equals" if not condition else condition
            if isinstance(value, datetime):
                if not value.hour and not value.minute:
                    self.value = str(value.date())
                else:
                    self.value = value.isoformat()
            else:
                self.value = str(value)

        if property_obj and not value:
            self.value = getattr(property_obj, "value", None)
            if isinstance(self.value, list):
                self.value = self.value[0]
        if self.condition in ["is_empty", "is_not_empty"]:
            self.value = True
        elif self.condition in ["past_week", "past_month", "past_year", "next_week", "next_month", "next_year"]:
            self.value = {}
        self.filter = {
            "property": self.property_name,
            self.property_type: {self.condition: self.value}
        }

    @property
    def allowed_condition_types(self):
        return ", ".join(self._filter_condition_types)

    def __repr__(self):
        if not getattr(self, "property_type"):
            return f"Filter({str(self.filter)})"
        return f"Filter({self.property_name} {self.condition} {self.value})"


class Sort(object):
    directions = ["ascending", "descending"]

    def __init__(self, property_name: str, direction: str):
        if direction not in self.directions:
            raise ValueError(f"Allowed types {self.directions} ({direction} is provided)")
        self.sorts = [{"property": property_name, "direction": direction}]

    def add(self, property_name: str, direction: str):
        if direction not in self.directions:
            raise ValueError(f"Allowed types {self.directions} ({direction} is provided)")
        self.sorts.append({"property": property_name, "direction": direction})

    def __repr__(self):
        r = [e.values() for e in self.sorts]
        return f"Sorts({r})"


class Request(object):
    def __init__(
            self,
            method: Optional[str] = None,
            path: Optional[str] = None,
            id_: str = "",
            data: Optional[Dict] = None,
            base: Optional[str] = None,
            token: Optional[str] = None,
            after_path: Optional[str] = None,
            limit: int = 0,
            filter_: Optional[Filter] = None,
            sorts: Optional[Sort] = None,
    ):
        self.session = requests.Session()
        self.base = base if base else envs.NOTION_URL
        self._token = token if token else envs.NOTION_SECRET
        if not self._token:
            logger.error("Token is not provided or file `token` is not found!")
        self.version = envs.NOTION_VERSION
        self.auth = {"Authorization": "Bearer " + self._token}
        self.headers = {"Notion-Version": self.version, **self.auth}
        self.result = None

        if method:
            self.result = self.method(method, path, id_, data, after_path, limit, filter_, sorts)

    def method(
            self, method, path, id_="", data=None, after_path=None,
            limit=0, filter_: Optional[Filter] = None, sorts: Optional[Sort] = None, pagination_loop: bool = False,
    ):
        if filter_:
            if data:
                data["filter"] = filter_.filter
            else:
                data = {"filter": filter_.filter}
        if sorts:
            if data:
                data["sorts"] = sorts.sorts
            else:
                data = {"sorts": sorts.sorts}
        url = self.base + path + "/" + id_
        if limit and method == "get":
            if after_path:
                after_path += "?" + urlencode({"page_size": limit})
            else:
                path += "?" + urlencode({"page_size": limit})
        if limit and method == "post":
            if not data:
                data = {}
            data.update({"page_size": limit})
        if after_path:
            url += "/" + after_path
        logger.info(f"Request {method} {url}")
        logger.debug(f"METHOD: {method.upper()}")
        logger.debug(f"URL: {url}")
        logger.debug(f"DATA: {data}")
        result = self.session.request(method=method, url=url, headers=self.headers, json=data)
        logger.debug(f"STATUS CODE: {result.status_code}")
        logger.debug(f"CONTENT: {result.content}")
        logger.info(f"{result.status_code} Received")

        r = find_response_error(result)

        # pagination section
        if not limit and not pagination_loop:
            self.paginate(r, method, path, id_, data, after_path)

        return r

    def paginate(self, result, method, path, id_, data, after_path):
        if (result.get("has_more", False) is True) and (result.get("object", "") == "list"):
            next_start = result.get("next_cursor")
            logger.info(f"Paginated answer. Repeat with offset {next_start}")

            super_after_path = after_path
            super_path = path

            while next_start:
                # if GET method then parameters are in request string
                # if POST method then parameters are in body
                if method == "get":
                    if after_path:
                        super_after_path = after_path + "?" + urlencode({"start_cursor": next_start})
                    else:
                        super_path = path + "?" + urlencode({"start_cursor": next_start})
                elif method == "post":
                    if not data:
                        data = {}
                    data.update({"start_cursor": next_start})

                r = self.method(method, super_path, id_, data, super_after_path, pagination_loop=True)
                if r.get("object", "") == "list" and r.get("results"):
                    result["results"].extend(r["results"])
                if r.get("has_more"):
                    next_start = r.get("next_cursor")
                else:
                    next_start = None



# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
from typing import Optional, Union, Dict, List

import pytion.envs as envs
from pytion.query import Request, Filter, Sort
from pytion.models import Database, Page, Block, BlockArray, PropertyValue, PageArray, LinkTo, RichTextArray, Property
from pytion.models import ElementArray, User


Models = Union[Database, Page, Block, BlockArray, PropertyValue, PageArray]
logger = logging.getLogger(__name__)


class Notion(object):
    def __init__(self, token: Optional[str] = None):
        self.version = envs.NOTION_VERSION
        self.session = Request(token=token)
        logger.debug(f"API object created. Version {envs.NOTION_VERSION}")

    def __len__(self):
        return 1

    # def __getstate__(self):
    #     return {"api": self.session}
    #
    # def __setstate__(self, d):
    #     self.__dict__.update(d)

    def __repr__(self):
        return "NotionAPI"

    def __str__(self):
        return self.__repr__()

    def __getattr__(self, name):
        return Element(self, name)


class Element(object):
    class_map = {"page": Page, "database": Database, "block": Block, "user": User}

    def __init__(self, api: Notion, name: str, obj: Optional[Models] = None):
        self.api = api
        self.name = name
        self.obj = obj
        logger.debug(f"Element {self!r} created")

    def get(self, id_: str, _after_path: str = None, limit: int = 0) -> Element:
        """
        Get Element by ID.
        .query.RequestError exception if not found

        :return:    `Element.obj` may be `Page`, `Database`, `Block`

        result = no.databases.get("1234123412341")
        result = no.pages.get("123412341234")
        result = no.blocks.get("123412341234")
        result = no.users.get("123412341234")
        print(result.obj)
        """
        if "-" in id_:
            id_ = id_.replace("-", "")
        if not _after_path:
            raw_obj = self.api.session.method(method="get", path=self.name, id_=id_, limit=limit)
        else:
            raw_obj = self.api.session.method(
                method="get", path=self.name, id_=id_, after_path=_after_path, limit=limit
            )
        if raw_obj["object"] == "list":
            if self.name == "pages":
                self.obj = PageArray(raw_obj["results"])
            elif self.name == "blocks":
                self.obj = BlockArray(raw_obj["results"])
            else:
                self.obj = ElementArray(raw_obj["results"])
        else:
            self.obj = self.class_map[raw_obj["object"]](**raw_obj)
        return self

    def get_parent(self, id_: Optional[str] = None) -> Optional[Element]:
        """
        Get parent object of current object if possible.

        :param id_:
        :return:    Element (parent object) or None. `.obj` may be `Page`, `Database`, `Block`

        `result = no.blocks.get_parent("123412341234")`
        `print(result)`
        Notion/pages/Page(Page Title here)
        `result = result.get_parent()`
        `print(result)`
        Notion/databases/Database(Some database name)
        """
        if not self.obj:
            self.get(id_)
        if getattr(self.obj, "parent", None):
            return self.from_linkto(self.obj.parent)
        logger.warning(f"Parent object can not be found")
        return None

    def get_block_children(
            self, id_: Optional[str] = None, block: Optional[Block] = None, limit: int = 0
    ) -> Optional[Element]:
        """
        Get children Block objects of current Block object (tabulated texts) if exist (else None)

        :param id_:
        :param block: you can provide a Block object instead to get his children
        :param limit:   0 < int < 100 - max number of items to be returned (0 = return all)
        :return:        `Element.obj` will be BlockArray object even nothing is found

        `print(no.pages.get_block_children("PAGE ID"))`
        None

        `print(no.blocks.get_block_children("PAGE ID"))`
        Notion/blocks/BlockArray(Heading 2 level Paragraph some)

        `print(no.blocks.get_block_children("PAGE ID").obj)`
        Heading 2 level
        Paragraph
        some text

        BlockArray or Database object expected.
        """
        if self.name not in ("blocks", "pages"):
            logger.warning("Only `blocks` or `pages` can have children")
            return None
        if isinstance(id_, str) and "-" in id_:
            id_ = id_.replace("-", "")
        obj = block if block else self.obj
        if obj:
            return self.from_linkto(obj.children, limit=limit)
        child = self.api.session.method(
            method="get", path="blocks", id_=id_, after_path="children", limit=limit
        )
        # children object returns list of Blocks
        if child["object"] != "list":
            logger.warning(f"List of Blocks expected. Received\n{child}")
            return None
        return Element(api=self.api, name="blocks", obj=BlockArray(child["results"]))

    def get_block_children_recursive(
        self, id_: Optional[str] = None, max_depth: int = 10, block: Optional[Block] = None,
        _cur_depth: int = 0, limit: int = 0, force: bool = False
    ) -> Optional[Element]:
        """
        Get children Block objects of current Block object (tabulated texts) if exist (else None) recursive

        :param id_:
        :param block:       you can provide a Block object instead to get his children
        :param max_depth:   how deep use the recursion (block inside block inside block etc.)
        :param limit:       0 < int < 100 - max number of items to be returned (0 = return all)
        :param force:       get blocks in subpages too
        :return:            `Element.obj` will be BlockArray object even nothing is found

        `print(no.blocks.get_block_children_recursive("PAGE ID").obj)`
        Heading 2 level
        Paragraph
            block inside block
        some text
        """
        if self.name not in ("blocks", "pages"):
            logger.warning("Only `blocks` or `pages` can have children")
            return None
        if isinstance(id_, str) and "-" in id_:
            id_ = id_.replace("-", "")
        obj = block if block else self.obj
        if obj:
            id_ = obj.id
            if isinstance(obj, Block) and obj.type == "child_database":
                return self.from_linkto(obj.children)
        child = self.api.session.method(
            method="get", path="blocks", id_=id_, after_path="children", limit=limit
        )
        ba = BlockArray([])
        for b in child["results"]:
            block_obj = Block(level=_cur_depth, **b)
            ba.append(block_obj)
            # Do not get subpages if not force
            if block_obj.type == "child_page" and not force:
                continue
            if block_obj.has_children and _cur_depth < max_depth:
                sub_element = Element(api=self.api, name="blocks").get_block_children_recursive(
                    id_=block_obj.id, max_depth=max_depth, _cur_depth=_cur_depth + 1, limit=limit, force=force
                )
                ba.extend(sub_element.obj)

        return Element(api=self.api, name="blocks", obj=ba)

    def get_page_property(self, property_id: str, id_: Optional[str] = None, limit: int = 0) -> Optional[Element]:
        """
        Retrieve a page property item.

        :param property_id: ID of property in current database
        :param id_:         ID of page in that database (if not `self.obj`)
        :param limit:       0 < int < 100 - max number of items to be returned (0 = return all)
        :return:            `Element.obj` will be PropertyValue object

        `db = no.databases.get("1232412341234")`
        `property_id = db.obj.properties["Last edited time"].id`
        `result = no.pages.get_page_property(property_id, 'PAGE ID 152f123a12344')`
        `print(result.obj)`
        2021-11-04 16:47:00+00:00
        """
        if self.name != "pages":
            logger.warning("Only `pages` can have properties")
            return None
        if isinstance(id_, str) and "-" in id_:
            id_ = id_.replace("-", "")
        if self.obj:
            id_ = self.obj.id
        property_obj = self.api.session.method(
            method="get", path=self.name, id_=id_, after_path="properties/"+property_id, limit=limit
        )
        return Element(api=self.api, name=f"pages/{id_}/properties", obj=PropertyValue(property_obj, property_id))

    def db_query(
            self,
            id_: Optional[str] = None,
            limit: int = 0,
            filter_: Optional[Filter] = None,
            sorts: Optional[Sort] = None,
            **kwargs,
    ) -> Optional[Element]:
        if self.name != "databases":
            logger.warning("Only `databases` can be queried")
            return None
        if isinstance(id_, str) and "-" in id_:
            id_ = id_.replace("-", "")
        if self.obj:
            id_ = self.obj.id
        r = self.api.session.method(
            method="post", path=self.name, id_=id_, after_path="query",
            data={}, limit=limit, filter_=filter_, sorts=sorts
        )
        if r["object"] != "list":
            return None
        return Element(api=self.api, name="pages", obj=PageArray(r["results"]))

    def db_filter(self, title: str = None, **kwargs) -> Optional[Element]:
        """
        :param title: filter by title contains + opt. attrs: condition, sort etc.
        OR
        :param property_name: mandatory - full name or ID of property to filter by
        :param value: the value of this property to filter by (may be bool or datetime etc.)
        :param property_type: mandatory field - `text`, `number`, `checkbox`, `date`, `select` etc.
        :param condition: optional field - it depends on the type: `starts_with`, `contains`, `equals` etc.
        :param raw: correctly formatted dict to pass direct to API (instead of all other params)
        :param property_obj: Property or PropertyValue obj. instead of `property_name` and `property_type`,
                             PropertyValue can put value in request, if `value` is not provided

        :param ascending: property name to be sorted by
        :param descending: property name to be sorted by

        :param limit: 0 < int < 100 - max number of items to be returned (0 = return all)
        :return:              self.obj -> PageArray

        examples
        `.db_filter("My Page Title")`
        `.db_filter("", ascending="Tags")`
        `.db_filter(property_name="Done", property_type="checkbox")`
        `.db_filter(property_name="Done", property_type="checkbox", value=False, descending="title")`
        `.db_filter(property_name="tags", property_type="multi_select", condition="is_not_empty")`
        `.db_filter(raw=YOUR_BIG_DICT_FROM_NOTION_DOCS, limit=2)`

        Filters combinations are not supported. (in `raw` param only)
        """
        if self.name == "databases" and self.obj:
            sort = None
            if kwargs.get("ascending"):
                sort = Sort(property_name=kwargs["ascending"], direction="ascending")
            elif kwargs.get("descending"):
                sort = Sort(property_name=kwargs["descending"], direction="descending")
            if isinstance(title, str):
                filter_obj = Filter(property_name="title", value=title, property_type="title", **kwargs)
            else:
                filter_obj = Filter(**kwargs)
            return self.db_query(filter_=filter_obj, sorts=sort, **kwargs)
        logger.warning("Database must be provided. use .get() before")
        return None

    def db_create(
            self, database_obj: Optional[Database] = None, parent: Optional[LinkTo] = None,
            properties: Optional[Dict[str, Property]] = None, title: Optional[Union[str, RichTextArray]] = None
    ) -> Optional[Element]:
        """
        :param database_obj:  you can provide `Database` object or -
                              provide the params for creating it:
        :param parent:
        :param properties:    dict of properties. Property with `title` type is mandatory!
        :param title:         your name of the Database
        :return:              self.obj -> Database

        `parent = LinkTo.create(database_id="24512345125123421")`
        `p1 = Property.create(name="renamed")`
        `p2 = Property.create(type_="multi_select", name="multiselected")`
        `props = {"Property1_name": p1, "Property2_name": p2}` OR
        ```props = {
            "Name": Property.create("title")
            "Digit": Property.create("number"),
            "Status": Property.create("select")
        }```
        `db = db.db_create(parent=parent, properties=props, title=RichTextArray.create("NEW DB"))`
        """
        if self.name != "databases":
            logger.warning("Method supports `databases` only")
            return None
        if database_obj:
            db = database_obj
        else:
            if isinstance(title, str):
                title = RichTextArray.create(title)
            db = Database.create(parent=parent, properties=properties, title=title)
        created_db = self.api.session.method(method="post", path=self.name, data=db.get())
        self.obj = Database(**created_db)
        return self

    def db_update(
            self, id_: Optional[str] = None, title: Optional[Union[str, RichTextArray]] = None,
            properties: Optional[Dict[str, Property]] = None
    ) -> Optional[Element]:
        """
        :param id_:         provide id of database if `self.obj` is empty
        :param title:       provide RichTextArray or text to rename database
        :param properties:  provide dict of Property to update them
        :return:            self.obj -> Database


        `rename_prop = Property.create(name="renamed")`
        `rename_retype_prop = Property.create(type_="multi_select", name="multiselected")`
        `retype_prop = Property.create("checkbox")`
        `props = {"Property1_name": rename_retype_prop, "Property2_ID": retype_prop}`
        `db = db.db_update(properties=props, title=RichTextArray.create("NEW DB"))`
        """
        if self.name != "databases":
            logger.warning("Method supports `databases` only")
            return None
        if isinstance(id_, str) and "-" in id_:
            id_ = id_.replace("-", "")
        if self.obj:
            id_ = self.obj.id
        patch = {}
        if title:
            if isinstance(title, str):
                title = RichTextArray.create(title)
            patch["title"] = title.get()
        if properties:
            patch["properties"] = {name: value.get() for name, value in properties.items()}
        updated_db = self.api.session.method(method="patch", path=self.name, id_=id_, data=patch)
        self.obj = Database(**updated_db)
        return self

    def page_create(
            self,
            page_obj: Optional[Page] = None,
            parent: Optional[LinkTo] = None,
            properties: Optional[Dict[str, PropertyValue]] = None,
            title: Optional[Union[str, RichTextArray]] = None,
            children: Optional[BlockArray, List[Block]] = None,
    ) -> Optional[Element]:
        """
        :param page_obj:      you can provide `Page` object or -
                              provide the params for creating it:
        :param parent:        LinkTo object with ID of parent element
        :param properties:    Dict of properties with values
        :param title:         New title
        :param children:      Content of new page in [Block] or BlockArray format
        :return:              self.obj -> Page

        `parent = LinkTo.create(database_id="24512345125123421")`
        `p2 = PropertyValue.create("date", datetime.now())`
        `r = no.pages.page_create(parent=parent, properties={"Count": p1, "Date": p2}, title="Extra PAGE")`

        `props["Status"] = PropertyValue.create("select", "new select option")`
        `props["Tags"] = PropertyValue.create("multi_select", ["new-option1", "new option2"])`
        `no.pages.create(parent=parent, properties=props)`

        `parent2 = LinkTo.create(page_id="123412341234")`
        `no.pages.page_create(parent=parent2, title="New page 121")`
        """
        if self.name != "pages":
            logger.warning("Method supports `pages` only")
            return None
        if page_obj:
            page = page_obj
        else:
            if children and not isinstance(children, BlockArray):
                children = BlockArray(children, create=True)
            page = Page.create(parent=parent, properties=properties, title=title, children=children)
        created_page = self.api.session.method(method="post", path=self.name, data=page.get())
        self.obj = Page(**created_page)
        return self

    def page_update(
            self, id_: Optional[str] = None, properties: Optional[Dict[str, PropertyValue]] = None,
            title: Optional[Union[str, RichTextArray]] = None, archived: bool = False
    ) -> Optional[Element]:
        """
        :param id_:         ID of page
        :param properties:  dict of existing properties
        :param title:
        :param archived:    set to `True` for delete the page
        :return:            self.obj -> Page
        """
        if self.name != "pages":
            logger.warning("Method supports `pages` only")
            return None
        if isinstance(id_, str) and "-" in id_:
            id_ = id_.replace("-", "")
        if self.obj:
            id_ = self.obj.id
        patch = {}
        if properties:
            patch["properties"] = {name: p.get() for name, p in properties.items()}
        if title:
            patch.setdefault("properties", {})
            patch["properties"]["title"] = PropertyValue.create("title", title).get()
        # if archived:
        patch["archived"] = archived
        updated_page = self.api.session.method(method="patch", path=self.name, id_=id_, data=patch)
        self.obj = Page(**updated_page)
        return self

    def block_update(
            self, id_: Optional[str] = None, block_obj: Optional[Block] = None,
            new_text: Optional[str] = None, archived: bool = False
    ) -> Optional[Element]:
        """
        Updates text of Block.
        `text`, `checked` (`to_do` type), `language` (`code` type) fields support only!
        You can modify any attrs of existing block and provide it (Block object) to this func.
        Changing the Block type is not supported.

        :param id_:         ID of block to change text OR
        :param block_obj:   modified Block (replace mode only)

        :param new_text:    new text (replace mode only)
        :param archived:    flag to delete that Block
        :return:            self.obj -> Block

        `blocks = no.blocks.get_block_children("PAGE ID")`
        `for b in blocks.obj:`
            `no.blocks.block_update(block_obj=b, new_text="OH YEEEAHH")`
        `for b in blocks.obj:`
            `b.text = "ALL IS DONE"`
            `no.blocks.block_update(block_obj=b)`
        """
        if self.name != "blocks":
            logger.warning("Method supports `blocks` only")
            return None
        if isinstance(id_, str) and "-" in id_:
            id_ = id_.replace("-", "")
        if block_obj:
            self.obj = block_obj
        if self.obj:
            id_ = self.obj.id
        else:
            self.get(id_)
        if not self.obj.get():
            return None
        if new_text:
            self.obj.text = new_text
        patch = {"archived": archived}
        patch.update(self.obj.get())
        updated_block = self.api.session.method(method="patch", path=self.name, id_=id_, data=patch)
        self.obj = Block(**updated_block)
        return self

    def block_append(
            self, id_: Optional[str] = None, block: Optional[Block] = None,
            blocks: Optional[BlockArray, List[Block]] = None
    ) -> Optional[Element]:
        """
        Append block or blocks children

        :param id_:         provide id of block or page if `self.obj` is empty

        :param block:       Block to append OR
        :param blocks:          List[Block] or BlockArray to append

        :return:            self.obj -> BlockArray

        `p1 = no.pages.get("PAGE ID")`
        `p1.block_append(block=Block.create("SOMETHING NEW YO"))`

        `no.blocks.block_append("BLOCK OR PAGE ID", blocks=blocks)`
        """
        if self.name not in ["blocks", "pages"]:
            logger.warning("Method supports `blocks` or `pages` only")
            return None
        if isinstance(id_, str) and "-" in id_:
            id_ = id_.replace("-", "")
        if self.obj:
            id_ = self.obj.id
        if isinstance(blocks, list):
            blocks = BlockArray(blocks, create=True)
        if isinstance(block, Block):
            blocks = BlockArray([block], create=True)
        data = {"children": blocks.get()}

        new_blocks = self.api.session.method(
            method="patch", path="blocks", id_=id_, after_path="children", data=data
        )
        return Element(api=self.api, name="blocks", obj=BlockArray(new_blocks["results"]))

    def from_linkto(self, linkto: LinkTo, limit: int = 0) -> Optional[Element]:
        if not linkto:
            logger.error("LinkTo must be provided!")
            return None
        if not linkto.uri:
            logger.error("LinkTo.uri must be provided!")
            return None
        new_element = Element(self.api, name=linkto.uri)
        return new_element.get(linkto.id, getattr(linkto, "after_path", None), limit)

    def from_object(self, model: Union[Database, Page, Block]):
        return Element(self.api, model.path, model)

    def __repr__(self):
        if not self.obj:
            return f"Notion/{self.name}/"
        return f"Notion/{self.name}/{self.obj!r}"

    def __str__(self):
        return self.__repr__()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytion",
    version="1.2rc1",
    author="Yegor Gomzin",
    author_email="slezycmex@mail.ru",
    description="Unofficial Python client for official Notion API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lastorel/pytion",
    project_urls={
        "Bug Tracker": "https://github.com/lastorel/pytion/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.7",
    install_requires=[
        "requests>=2.26.0"
    ]
)



# -*- coding: utf-8 -*-

import logging
import json
from typing import Dict

from requests import Response


logger = logging.getLogger(__name__)


class ClientError(Exception):
    def __init__(self, message: Response):
        req = message
        message = f"Failed with code {req.status_code}. Raw: {req.content}"
        super(ClientError, self).__init__(message)
        self.req = req
        self.request_body = req.request.body
        self.base = req.url
        self.error = message


class ServerError(Exception):
    def __init__(self, message: Response):
        req = message
        message = f"Failed with code {req.status_code}. Raw: {req.content}"
        super(ServerError, self).__init__(message)
        self.req = req
        self.request_body = req.request.body
        self.base = req.url
        self.error = message


class InvalidJSON(ClientError):
    def __init__(self, req: Response):
        message = f"Body could not be decoded as JSON: {req.request.body}"
        Exception.__init__(self, message)


class InvalidRequestURL(ClientError):
    def __init__(self, req: Response):
        message = f"The request URL is not valid: {req.url}"
        Exception.__init__(self, message)


class InvalidRequest(ClientError):
    def __init__(self, req: Response):
        message = f"This request is not supported: {req.url} {req.request.method}"
        Exception.__init__(self, message)


class ValidationError(ClientError):
    def __init__(self, content: Dict):
        message = content.get("message")
        message = f"The request body does not match the schema: {message}"
        Exception.__init__(self, message)


class MissingVersion(ClientError):
    def __init__(self, *args):
        message = "The request is missing the required Notion-Version header."
        Exception.__init__(self, message)


class Unauthorized(ClientError):
    def __init__(self, *args):
        message = "The bearer token is not valid."
        Exception.__init__(self, message)


class RestrictedResource(ClientError):
    def __init__(self, *args):
        message = "Given the bearer token used, the client doesn't have permission to perform this operation."
        Exception.__init__(self, message)


class ObjectNotFound(ClientError):
    def __init__(self, req: Response):
        message = f"{req.url}" \
                  "The resource does not exist or the resource has not been shared with owner of the bearer token."
        Exception.__init__(self, message)


class ConflictError(ClientError):
    def __init__(self, *args):
        message = "The transaction could not be completed, potentially due to a data collision. Make sure the " \
                  "parameters are up to date and try again."
        Exception.__init__(self, message)


class RateLimited(ClientError):
    def __init__(self, *args):
        message = "This request exceeds the number of requests allowed. Slow down and try again."
        Exception.__init__(self, message)


class InternalServerError(ServerError):
    def __init__(self, req: Response):
        self.req = req
        self.request_body = req.request.body
        self.base = req.url
        self.error = req.content
        message = f"An unexpected error occurred. Reach out to Notion support. {req.status_code}: " \
                  f"{req.request.method} {self.base} {self.request_body} -> {self.error}"
        Exception.__init__(self, message)


class ServiceUnavailable(ServerError):
    def __init__(self, *args):
        message = "Notion is unavailable. Try again later."
        Exception.__init__(self, message)


class DatabaseConnectionUnavailable(ServerError):
    def __init__(self, *args):
        message = "Notion's database is unavailable or in an unqueryable state. Try again later."
        Exception.__init__(self, message)


class ContentError(Exception):
    """Content Exception
    If the API URL does not point to a valid Notion API, the server may
    return a valid response code, but the content is not json. This
    exception is raised in those cases.
    """

    def __init__(self, message):
        req = message

        message = (
            "The server returned invalid (non-json) data. Maybe not a Notion server?"
        )

        super(ContentError, self).__init__(message)
        self.req = req
        self.request_body = req.request.body
        self.base = req.url
        self.error = message


def find_response_error(req: Response) -> Dict:
    try:
        content = req.json()
    except json.JSONDecodeError:
        logger.error(f"Result is not OK. JSON decoding fail\n{req.content}")
        raise ContentError(req)
    if req.ok:
        return content
    logger.error(f"Result is not OK. {req.status_code}\n{req.reason}")
    status_code = int(req.status_code)
    error_code = content.get("code")
    if error_code:
        if error_code == "invalid_json":
            raise InvalidJSON(req)
        elif error_code == "invalid_request_url":
            raise InvalidRequestURL(req)
        elif error_code == "invalid_request":
            raise InvalidRequest(req)
        elif error_code == "validation_error":
            raise ValidationError(content)
        elif error_code == "missing_version":
            raise MissingVersion()
        elif error_code == "unauthorized":
            raise Unauthorized()
        elif error_code == "restricted_resource":
            raise RestrictedResource()
        elif error_code == "object_not_found":
            raise ObjectNotFound(req)
        elif error_code == "conflict_error":
            raise ConflictError()
        elif error_code == "rate_limited":
            raise RateLimited()
        elif error_code == "internal_server_error":
            raise InternalServerError(req)
        elif error_code == "service_unavailable":
            raise ServiceUnavailable()
        elif error_code == "database_connection_unavailable":
            raise DatabaseConnectionUnavailable()
    if 400 <= status_code < 500:
        raise ClientError(req)
    elif 500 <= status_code < 600:
        raise ServerError(req)
    raise Exception(f"An unexpected error occurred. {req.status_code}: {req.content}")
